// src/babel-esm.ts
import { parse } from "@babel/parser";
import * as t from "@babel/types";
import { createRequire } from "module";
var require2 = createRequire(import.meta.url);
var _traverse = require2("@babel/traverse");
var traverse = _traverse.default;

// src/identifier.ts
function fromFunction(path) {
  let parentPath = path.parentPath;
  if (parentPath.type === "VariableDeclarator") {
    let variablePath = parentPath;
    let name = variablePath.get("id");
    return name.node.type === "Identifier" ? name : null;
  }
  if (parentPath.type === "AssignmentExpression") {
    let variablePath = parentPath;
    let name = variablePath.get("left");
    return name.node.type === "Identifier" ? name : null;
  }
  if (path.node.type === "ArrowFunctionExpression") {
    return null;
  }
  return path.node.id && path.node.id.type === "Identifier" ? path.get("id") : null;
}
function isReferenced(ident) {
  let binding = ident.scope.getBinding(ident.node.name);
  if (binding?.referenced) {
    if (binding.path.type === "FunctionDeclaration") {
      return !binding.constantViolations.concat(binding.referencePaths).every((ref) => ref.findParent((parent) => parent === binding?.path));
    }
    return true;
  }
  return false;
}

// src/dead-code-elimination.ts
function dead_code_elimination_default(ast, candidates) {
  let referencesRemovedInThisPass;
  let shouldBeRemoved = (ident) => {
    if (isReferenced(ident))
      return false;
    if (!candidates)
      return true;
    return candidates.has(ident);
  };
  let sweepFunction = (path) => {
    let identifier = fromFunction(path);
    if (identifier?.node && shouldBeRemoved(identifier)) {
      ++referencesRemovedInThisPass;
      if (t.isAssignmentExpression(path.parentPath.node) || t.isVariableDeclarator(path.parentPath.node)) {
        path.parentPath.remove();
      } else {
        path.remove();
      }
    }
  };
  let sweepImport = (path) => {
    let local = path.get("local");
    if (shouldBeRemoved(local)) {
      ++referencesRemovedInThisPass;
      path.remove();
      if (path.parent.specifiers.length === 0) {
        path.parentPath.remove();
      }
    }
  };
  do {
    referencesRemovedInThisPass = 0;
    traverse(ast, {
      Program(path) {
        path.scope.crawl();
      },
      // eslint-disable-next-line no-loop-func
      VariableDeclarator(path) {
        let id = path.get("id");
        if (id.isIdentifier()) {
          if (shouldBeRemoved(id)) {
            ++referencesRemovedInThisPass;
            path.remove();
          }
        } else if (id.isObjectPattern() || id.isArrayPattern()) {
          let vars = findVariablesInPattern(id);
          for (let v of vars) {
            if (isReferenced(v))
              continue;
            let parent = v.parentPath;
            ++referencesRemovedInThisPass;
            if (parent.isObjectProperty()) {
              parent.remove();
              continue;
            }
            if (parent.isArrayPattern()) {
              parent.node.elements[v.key] = null;
              continue;
            }
            if (parent.isAssignmentPattern()) {
              if (t.isObjectProperty(parent.parent)) {
                parent.parentPath?.remove();
                continue;
              }
              if (t.isArrayPattern(parent.parent)) {
                parent.parent.elements[parent.key] = null;
                continue;
              }
              throw unexpected(parent);
            }
            if (parent.isRestElement()) {
              parent.remove();
              continue;
            }
            throw unexpected(parent);
          }
        }
      },
      ObjectPattern(path) {
        let isWithinDeclarator = path.find((p) => p.isVariableDeclarator()) !== null;
        let isEmpty = path.node.properties.length === 0;
        if (isWithinDeclarator && isEmpty) {
          removePattern(path);
          referencesRemovedInThisPass++;
        }
      },
      ArrayPattern(path) {
        let isWithinDeclarator = path.find((p) => p.isVariableDeclarator()) !== null;
        let isEmpty = path.node.elements.every((e) => e === null);
        if (isWithinDeclarator && isEmpty) {
          removePattern(path);
          referencesRemovedInThisPass++;
        }
      },
      FunctionDeclaration: sweepFunction,
      FunctionExpression: sweepFunction,
      ArrowFunctionExpression: sweepFunction,
      ImportSpecifier: sweepImport,
      ImportDefaultSpecifier: sweepImport,
      ImportNamespaceSpecifier: sweepImport
    });
  } while (referencesRemovedInThisPass);
}
function findVariablesInPattern(patternPath) {
  let variables = [];
  function recurse(path) {
    if (path.isIdentifier()) {
      variables.push(path);
      return;
    }
    if (path.isObjectPattern()) {
      return path.get("properties").forEach(recurse);
    }
    if (path.isObjectProperty()) {
      return recurse(path.get("value"));
    }
    if (path.isArrayPattern()) {
      let _elements = path.get("elements");
      return _elements.forEach(recurse);
    }
    if (path.isAssignmentPattern()) {
      return recurse(path.get("left"));
    }
    if (path.isRestElement()) {
      return recurse(path.get("argument"));
    }
    if (path.node === null)
      return;
    throw unexpected(path);
  }
  recurse(patternPath);
  return variables;
}
function removePattern(path) {
  let parent = path.parentPath;
  if (parent.isVariableDeclarator()) {
    return parent.remove();
  }
  if (parent.isArrayPattern()) {
    parent.node.elements[path.key] = null;
    return;
  }
  if (parent.isObjectProperty()) {
    return parent.remove();
  }
  if (parent.isRestElement()) {
    return parent.remove();
  }
  if (parent.isAssignmentPattern()) {
    if (t.isObjectProperty(parent.parent)) {
      return parent.parentPath.remove();
    }
    if (t.isArrayPattern(parent.parent)) {
      parent.parent.elements[parent.key] = null;
      return;
    }
    throw unexpected(parent.parentPath);
  }
  throw unexpected(parent);
}
function unexpected(path) {
  let type = path.node === null ? "null" : path.node.type;
  return path.buildCodeFrameError(
    `[babel-dead-code-elimination] unexpected node type: ${type}`
  );
}

// src/find-referenced-identifiers.ts
function find_referenced_identifiers_default(ast) {
  const referenced = /* @__PURE__ */ new Set();
  function markFunction(path) {
    const ident = fromFunction(path);
    if (ident?.node && isReferenced(ident)) {
      referenced.add(ident);
    }
  }
  function markImport(path) {
    const local = path.get("local");
    if (isReferenced(local)) {
      referenced.add(local);
    }
  }
  traverse(ast, {
    VariableDeclarator(path) {
      if (path.node.id.type === "Identifier") {
        const local = path.get("id");
        if (isReferenced(local)) {
          referenced.add(local);
        }
      } else if (path.node.id.type === "ObjectPattern") {
        const pattern = path.get("id");
        const properties = pattern.get("properties");
        properties.forEach((p) => {
          const local = p.get(
            p.node.type === "ObjectProperty" ? "value" : p.node.type === "RestElement" ? "argument" : function() {
              throw new Error("invariant");
            }()
          );
          if (isReferenced(local)) {
            referenced.add(local);
          }
        });
      } else if (path.node.id.type === "ArrayPattern") {
        const pattern = path.get("id");
        const elements = pattern.get("elements");
        elements.forEach((e) => {
          let local;
          if (e.node?.type === "Identifier") {
            local = e;
          } else if (e.node?.type === "RestElement") {
            local = e.get("argument");
          } else {
            return;
          }
          if (isReferenced(local)) {
            referenced.add(local);
          }
        });
      }
    },
    FunctionDeclaration: markFunction,
    FunctionExpression: markFunction,
    ArrowFunctionExpression: markFunction,
    ImportSpecifier: markImport,
    ImportDefaultSpecifier: markImport,
    ImportNamespaceSpecifier: markImport
  });
  return referenced;
}
export {
  dead_code_elimination_default as deadCodeElimination,
  find_referenced_identifiers_default as findReferencedIdentifiers
};
